@@ -1,9 +1,9 @@
-from typing import TYPE_CHECKING, Iterator
+from typing import TYPE_CHECKING, Any, Iterator
 
 from typing_extensions import Protocol
 
 if TYPE_CHECKING:
-    from .fs.base import AnyFSPath, FileSystem
+    from dvc_objects.fs.base import AnyFSPath, FileSystem
 
 # pylint: disable=unused-argument
 
@@ -13,3 +13,6 @@ class Ignore(Protocol):
         self, fs: "FileSystem", path: "AnyFSPath"
     ) -> Iterator["AnyFSPath"]:
         ...
+
+    def walk(self, fs: "FileSystem", path: "AnyFSPath", **kwargs: Any):
+        ...
